use std::{ffi::OsString, os::fd::{AsRawFd, OwnedFd}};
use super::super::uapi;
use std::path::Path;
use std::fs::File;

#[derive(Debug)]
pub struct DrmDeviceNode {
    pub fd: OwnedFd,
    pub path: Option<OsString>,
}

#[derive(Debug)]
pub struct Device {
    pub node: DrmDeviceNode,
    pub driver: uapi::DrmVersion,
}

#[derive(Debug)]
pub struct Engine {
    pub info: uapi::i915::EngineInfo,
}

pub fn find_device() -> Option<Device> {
    uapi::i915::find_node().and_then(|file| Device::from_path(file.path?))
}

impl Device {
    pub fn from_fd(fd: OwnedFd) -> Option<Device> {
        let driver = uapi::get_drm_version(fd.as_raw_fd())?;
        Some(Device {
            node: DrmDeviceNode {
                fd: fd,
                path: None,
            },
            driver: driver,
        })
    }
    pub fn from_path(path: OsString) -> Option<Device> {
        let file = File::open(Path::new(&path)).ok()?;
        let driver = uapi::get_drm_version(file.as_raw_fd())?;
        Some(Device {
            node: DrmDeviceNode {
                fd: OwnedFd::from(file),
                path: Some(path),
            },
            driver: driver,
        })
    }
    pub fn get_engines(&self) -> Option<Vec<Engine>> {
        uapi::i915::get_engines(self.node.fd.as_raw_fd()).map(|engines| {
            engines.into_iter().map(|engine_info| {
                Engine {
                    info: engine_info,
                }
            }).collect()
        })
    }
    pub fn get_param(&self, param: u32) -> Option<i32> {
        uapi::i915::get_param(self.node.fd.as_raw_fd(), param as i32)
    }
}